// SPDX-FileCopyrightText: 2021-2025 The Ikarus Developers ikarus@ibb.uni-stuttgart.de
// SPDX-License-Identifier: LGPL-3.0-or-later

/**
 * \file easfunctions.hh
 * \brief Header file for various EAS functions.
 * \ingroup eas
 */

#pragma once

#include <ikarus/finiteelements/mechanics/strainenhancements/easfunctions/displacementgradient.hh>
#include <ikarus/finiteelements/mechanics/strainenhancements/easfunctions/displacementgradienttransposed.hh>
#include <ikarus/finiteelements/mechanics/strainenhancements/easfunctions/greenlagrangestrain.hh>
#include <ikarus/finiteelements/mechanics/strainenhancements/easfunctions/linearstrain.hh>
